# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from EvoMSA.command_line import train
from EvoMSA.command_line import predict
from microtc.utils import load_model, tweet_iterator
import sys
import os
from test_base import TWEETS
from nose.tools import assert_almost_equals


def test_train():
    from EvoMSA.base import EvoMSA
    sys.argv = ['EvoMSA', '-ot.model', '-n2',
                '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                TWEETS]
    train(output=True)
    evo = load_model('t.model')
    assert isinstance(evo, EvoMSA)
    os.unlink('t.model')


def test_evo_kwargs():
    from EvoMSA.base import EvoMSA
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}, "b4msa_args": {"del_dup":false}}',
                '-ot.model',
                '-n2', TWEETS]
    train(output=True)
    evo = load_model('t.model')
    assert isinstance(evo, EvoMSA)
    os.unlink('t.model')


def test_predict():
    import numpy as np
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                '-ot.model', '-n2', TWEETS]
    train(output=True)
    sys.argv = ['EvoMSA', '-mt.model', '-ot1.json', TWEETS]
    predict()
    hy = np.array([x['klass'] for x in tweet_iterator('t1.json')])
    [x['decision_function'] for x in tweet_iterator('t1.json')]
    y = np.array([x['klass'] for x in tweet_iterator(TWEETS)])
    acc = (y == hy).mean()
    print(acc)
    assert acc <= 1 and acc > 0.8
    os.unlink('t1.json')
    os.unlink('t.model')


def test_evo_test_set():
    from EvoMSA.base import EvoMSA
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                '-ot.model', '--test_set', TWEETS, '-n2', TWEETS]
    train(output=True)
    evo = load_model('t.model')
    assert isinstance(evo, EvoMSA)
    os.unlink('t.model')


def test_raw_outputs():
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 60, "n_estimators": 30}}',
                '-ot.model', '-n2', TWEETS]
    train(output=True)
    sys.argv = ['EvoMSA', '--raw-outputs', '-mt.model', '-ot1.json', TWEETS]
    predict()
    df = [x['decision_function'] for x in tweet_iterator('t1.json')]
    assert len(df[0]) == 30 * 4
    os.unlink('t1.json')
    os.unlink('t.model')


def test_decision_function():
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                '-ot.model', '-n2', TWEETS]
    train(output=True)
    sys.argv = ['EvoMSA', '--decision-function', '-mt.model', '-ot1.json', TWEETS]
    predict()
    df = [x['decision_function'] for x in tweet_iterator('t1.json')]
    assert len(df[0]) == 4
    os.unlink('t1.json')
    os.unlink('t.model')


def test_max_lines():
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}', '-ot.model', '-n2', TWEETS]
    train()
    sys.argv = ['EvoMSA', '-mt.model', '--max-lines', '500', '-ot.json', TWEETS]
    predict()
    os.unlink('t.model')
    os.unlink('t.json')


def test_evo_test_set_shuffle():
    from EvoMSA.base import EvoMSA
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                '-ot.model', '--test_set', 'shuffle', '-n2', TWEETS]
    train(output=True)
    evo = load_model('t.model')
    assert isinstance(evo, EvoMSA)
    os.unlink('t.model')


def test_predict_numbers():
    from sklearn.preprocessing import LabelEncoder
    import json
    d = [x for x in tweet_iterator(TWEETS)]
    le = LabelEncoder().fit([x['klass'] for x in d])
    y = le.transform([x['klass'] for x in d]).tolist()
    with open('ex.json', 'w') as fpt:
        for x, y0 in zip(d, y):
            x['klass'] = y0
            fpt.write(json.dumps(x) + '\n')
    sys.argv = ['EvoMSA', '--kw={"stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}, "models": [["EvoMSA.model.Corpus", "EvoMSA.model.Bernulli"]], "TR": false}',
                '-ot.model', '-n1', 'ex.json']
    train(output=True)
    sys.argv = ['EvoMSA', '-mt.model', '-ot1.json', TWEETS]
    predict()


def test_predict_NearestCentroid():
    from sklearn.preprocessing import LabelEncoder
    import json
    d = [x for x in tweet_iterator(TWEETS)]
    le = LabelEncoder().fit([x['klass'] for x in d])
    y = le.transform([x['klass'] for x in d]).tolist()
    with open('ex.json', 'w') as fpt:
        for x, y0 in zip(d, y):
            x['klass'] = y0
            fpt.write(json.dumps(x) + '\n')
    sys.argv = ['EvoMSA',
                '--kw={"stacked_method": "sklearn.neighbors.NearestCentroid", "TR": false, "models": [["EvoMSA.model.Corpus", "EvoMSA.model.Bernulli"]]}',
                '-ot.model', '-n1', 'ex.json']
    train(output=True)
    sys.argv = ['EvoMSA', '-mt.model', '-ot1.json', TWEETS]
    predict()


def test_performance_validation_set():
    # import os
    from EvoMSA.command_line import performance, fitness_vs
    for seed in range(3):
        # if os.path.isfile('t-%s.model' % seed):
        #     continue
        sys.argv = ['EvoMSA', '--kw={"seed": %s, "stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "n_estimators": 5}}' % seed,
                    '-ot-%s.model' % seed, '-n1', TWEETS]
        train()
    sys.argv = ['EvoMSA', '-m'] + ['t-%s.model' % seed for seed in range(3)]
    print(fitness_vs((0, 't-0.model')))
    m = performance(output=True)
    assert len(m._p) == 3
    # assert False


def test_performance_validation_set2():
    import os
    from EvoMSA.command_line import performance
    for seed in range(4):
        print('haciendo', seed)
        if os.path.isfile('t-%s.model' % seed):
            continue
        sys.argv = ['EvoMSA', '--kw={"seed": %s, "stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 3}}' % seed,
                    '-ot-%s.model' % seed, '-n1', TWEETS]
        train()
    sys.argv = ['EvoMSA', '-n2', '-m'] + ['t-%s.model' % seed for seed in range(2)] + ['-'] + ['t-%s.model' % seed for seed in range(2, 4)]
    m = performance(output=True)
    assert len(m._p) == 2


def test_performance_public_set():
    import os
    from EvoMSA.command_line import performance
    for seed in range(4):
        if os.path.isfile('t-%s.model' % seed):
            continue
        sys.argv = ['EvoMSA', '--kw={"seed": %s, "stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 3}}' % seed,
                    '-ot-%s.model' % seed, '-n1', TWEETS]
        train(output=True)
    for seed in range(4):
        if os.path.isfile('t-%s.predict' % seed):
            continue
        sys.argv = ['EvoMSA', '-mt-%s.model' % seed, '-ot-%s.predict' % seed, TWEETS]
        predict()
    for score in ['macroF1', 'macroRecall', 'macroPrecision', 'accuracy']:
        sys.argv = ['EvoMSA', '--score', score, '-n2', '-y', TWEETS] + ['t-%s.predict' % seed for seed in range(2)] + ['-'] + ['t-%s.predict' % seed for seed in range(2, 4)]
        m = performance(output=True)
        assert len(m._p) == 2


def test_list_of_text():
    import os
    import json
    from EvoMSA.command_line import train
    with open('t.json', 'w') as fpt:
        for x in tweet_iterator(TWEETS):
            x['text'] = [x['text'], x['text']]
            fpt.write(json.dumps(x) + '\n')
    sys.argv = ['EvoMSA', '-ot.model', '-n2',
                '--kw={"models": [["EvoMSA.model.Corpus", "EvoMSA.model.Bernulli"]], "lang": "es", "TR": false, "stacked_method_args": {"popsize": 10, "early_stopping_rounds": 10, "time_limit": 5, "n_estimators": 5}}',
                't.json']
    train()
    os.unlink('t.json')




# Copyright 2018 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from b4msa.textmodel import TextModel
from microtc.utils import tweet_iterator
from scipy.sparse import csr_matrix
from sklearn.svm import LinearSVC
from ConceptModelling.thumbs_up_down import ThumbsUpDown, _ARABIC, _ENGLISH, _SPANISH, PATH as ConPATH
import os
from microtc.utils import save_model
from sklearn.neighbors import KDTree


class BaseTextModel(object):
    """Base class for text model

    :param corpus: Text to build the text model
    :type corpus: list or dict
    """

    def __init__(self, corpus=None, **kwargs):
        pass

    def fit(self, X):
        """
        Train the model

        :param X: Corpus
        :type X: list
        :rtype: instance
        """

        pass

    def __getitem__(self, x):
        pass

    def tokenize(self, text):
        pass

    def transform(self, X):
        return np.array([self.__getitem__(x) for x in X])


class BaseClassifier(object):
    """Base class for the classifier"""

    def __init__(self, random_state=0):
        pass

    def fit(self, X, y):

        """Method to train the classifier
        :param X: Independent variable
        :type X: np.array or csc_matrix
        :param y: Dependent variable
        :type y: np.array
        :rtype: self
        """

        return self

    def decision_function(self, X):
        """Classifier's decision function

        :param X: Independent variable
        :type X: np.array or csc_matrix
        :rtype: np.array
        """

        pass


class OutputClassifier(object):
    """LinearSVC that outputs the training set and test set using the environment varible OUTPUT"""

    def __init__(self, random_state=0, output=None):
        self._output = os.getenv('OUTPUT', output)
        assert self._output is not None
        self._random_state = random_state

    def fit(self, X, y):
        self.m = LinearSVC(random_state=self._random_state).fit(X, y)
        try:
            X = np.array(X.todense())
        except AttributeError:
            pass
        with open('%s_train.csv' % self._output, 'w') as fpt:
            for x, _y in zip(X, y):
                fpt.write(",".join([str(_) for _ in x]))
                fpt.write(",%s\n" % str(_y))
        return self

    def decision_function(self, X):
        hy = self.m.decision_function(X)
        try:
            X = np.array(X.todense())
        except AttributeError:
            pass
        with open('%s_test.csv' % self._output, 'w') as fpt:
            for x in X:
                fpt.write(",".join([str(_) for _ in x]))
                fpt.write("\n")
        return hy


class EvoMSAWrapper(BaseTextModel):
    def __init__(self, evomsa=None):
        assert evomsa is not None
        evomsa.n_jobs = 1
        self._evomsa = evomsa

    def transform(self, X):
        return self._evomsa.predict_proba(X)


class Identity(BaseTextModel, BaseClassifier):
    """Identity function used as either text model or classifier or regressor"""

    def __getitem__(self, x):
        return x

    def fit(self, X, y=None):
        return self

    def decision_function(self, X):
        try:
            return X.toarray()
        except AttributeError:
            return X

    def predict_proba(self, X):
        return self.decision_function(X)


class HA(BaseTextModel):
    """Wrapper of b4msa.textmodel.TextModel and LinearSVC"""
    def __init__(self, **kwargs):
        self._tm = TextModel(**kwargs)
        self._cl = LinearSVC()

    def fit(self, X, y):
        self._tm.fit(X)
        self._cl.fit(self._tm.transform(X), y)
        return self

    def transform(self, X):
        res = self._cl.decision_function(self._tm.transform(X))
        if res.ndim == 1:
            return np.atleast_2d(res).T
        return res

    @classmethod
    def create_space(cls, fname, output, **kwargs):
        """Create the model from a file of json

        :param fname: Path to the file containing the json
        :type fname: str
        :param output: Path to store the model
        :type output: str
        :param kwargs: Keywords pass to TextModel
        """

        X = [x for x in tweet_iterator(fname)]
        m = cls(**kwargs)
        m.fit(X, [x['klass'] for x in X])
        save_model(m, output)


class Projection(BaseTextModel):
    def __init__(self, docs=None, textModel=None, projection=None):
        assert docs is None
        self._textModel = textModel
        self._projection = projection

    def transform(self, X):
        return np.dot(self._textModel.transform(X), self._projection)

    def __getitem__(self, x):
        return np.dot(self._textModel[x], self._projection)


class LabeledDataSet(BaseTextModel, BaseClassifier):
    """Create a text classifier using b4msa.textmodel.TextModel and LinearSVC

    :param docs: do not use
    :type docs: None
    :param textModel: text model e.g., b4msa.textmodel.TextModel
    :param coef: coefficients obtained from LinearSVC
    :type coef: array
    :param intercept: bias obtained from LinearSVC
    :type intercept: array
    :param labels: list of labels or classes
    :type labels: list
    """
    def __init__(self, docs=None, textModel=None, coef=None, intercept=None, labels=None):
        assert docs is None
        self._textModel = textModel
        self._coef = coef
        self._intercept = intercept
        self._labels = labels
        self._text = os.getenv('TEXT', default='text')

    @property
    def textModel(self):
        from .cython_utils import TextModelPredict
        try:
            return self._textModelPredict
        except AttributeError:
            self._textModelPredict = TextModelPredict(self._textModel,
                                                      self._coef,
                                                      self._intercept)
        return self._textModelPredict

    def __getstate__(self):
        """Remove attributes before the pickle"""

        r = self.__dict__.copy()
        try:
            del r['_textModelPredict']
        except KeyError:
            pass
        return r

    def fit(self, X, y=None):
        return self

    def get_text(self, text):
        key = self._text
        if isinstance(text, (list, tuple)):
            return " | ".join([x[key] for x in text])
        elif isinstance(text, str):
            return text
        return text[key]

    def decision_function(self, X):
        return self.transform(X)

    def predict_proba(self, X):
        return self.decision_function(X)

    def transform(self, X):
        output = []
        self.textModel.transform([self.get_text(x) for x in X], output)
        return np.array(output)

    def __getitem__(self, x):
        return np.array(self.textModel[self.get_text(x)])

    @classmethod
    def _create_space(cls, fname, **kwargs):
        """Create the space from a file of json

        :param fname: Path to the file containing the json
        :type fname: str
        :param kwargs: Keywords pass to TextModel
        """
        import random
        from .utils import linearSVC_array
        from collections import Counter
        try:
            from tqdm import tqdm
        except ImportError:
            def tqdm(x, **kwargs):
                return x

        data = [x for x in tweet_iterator(fname)]
        random.shuffle(data)
        tm = TextModel(**kwargs).fit([x['text'] for x in data[:128000]])
        tm._num_terms = tm.model.num_terms
        # klass, nele = np.unique([x['klass'] for x in data], return_counts=True)
        _ = [(k, v) for k, v in Counter([x['klass'] for x in data]).items()]
        _.sort(key=lambda x: x[0])
        klass = [x[0] for x in _]
        nele = [x[1] for x in _]
        h = {v: k for k, v in enumerate(klass)}
        MODELS = []
        for ident, k in tqdm(enumerate(klass)):
            elepklass = [0 for __ in klass]
            cnt = nele[ident]
            cntpklass = int(cnt / (len(klass) - 1))
            D = [(x, 1) for x in data if x['klass'] == k]
            for x in data:
                if x['klass'] == k:
                    continue
                if elepklass[h[x['klass']]] > cntpklass:
                    continue
                elepklass[h[x['klass']]] = elepklass[h[x['klass']]] + 1
                D.append((x, -1))
            m = LinearSVC().fit(tm.tonp([tm[x[0]['text']] for x in D]), [x[1] for x in D])
            MODELS.append(m)
        coef, intercept = linearSVC_array(MODELS)
        return tm, coef, intercept, klass

    @classmethod
    def create_space(cls, fname, output=None, **kwargs):
        """Create the space from a file of json

        :param fname: Path to the file containing the json
        :type fname: str
        :param output: Path to store the model, it is cls.model_fname if None
        :type output: str
        :param kwargs: Keywords pass to TextModel
        """
        tm, coef, intercept, klass = cls._create_space(fname, **kwargs)
        if output is None:
            output = cls.model_fname()
        ins = cls(textModel=tm, coef=coef, intercept=intercept, labels=klass)
        save_model(ins, output)


class EmoSpace(LabeledDataSet):
    """Base class of Emoji Space. Use EmoSpaceAr or EmoSpaceEn or EmoSpaceEs instead of this class
    """

    def __init__(self, docs=None, model_cl=None, **kwargs):
        if model_cl is None:
            textModel, coef, intercept, labels = self.get_model()
        else:
            textModel, coef, intercept, labels = model_cl
        super(EmoSpace, self).__init__(textModel=textModel, coef=coef,
                                       intercept=intercept, labels=labels)

    def get_model(self):
        from .utils import get_model
        model_fname = self.model_fname()
        return get_model(model_fname)

    @classmethod
    def _create_space(cls, fname, emo_option='delete', **kwargs):
        """Create the space from a file of json

        :param fname: Path to the file containing the json
        :type fname: str
        :param kwargs: Keywords pass to TextModel
        """
        return super(EmoSpace, cls)._create_space(fname, emo_option=emo_option, **kwargs)

    @classmethod
    def create_space(cls, fname, output=None, **kwargs):
        """Create the space from a file of json

        :param fname: Path to the file containing the json
        :type fname: str
        :param output: Path to store the model, it is cls.model_fname if None
        :type output: str
        :param kwargs: Keywords pass to TextModel
        """
        tm, coef, intercept, klass = cls._create_space(fname, **kwargs)
        if output is None:
            output = cls.model_fname()
        save_model([tm, coef, intercept, klass], output)


class EmoSpaceEs(EmoSpace):
    """Spanish text model or classifier based on a Emoji Space
    """
    @staticmethod
    def model_fname():
        import EvoMSA
        return 'emo-v%s-es.evoemo' % EvoMSA.__version__[:3]

    @classmethod
    def create_space(cls, fname, output=None, lang='es', **kwargs):
        super(EmoSpaceEs, cls).create_space(fname, output=output, lang=lang, **kwargs)


class EmoSpaceEn(EmoSpace):
    """English text model or classifier based on a Emoji Space"""

    @staticmethod
    def model_fname():
        import EvoMSA
        return 'emo-v%s-en.evoemo' % EvoMSA.__version__[:3]

    @classmethod
    def create_space(cls, fname, output=None, lang='en', **kwargs):
        super(EmoSpaceEn, cls).create_space(fname, output=output, lang=lang, **kwargs)


class EmoSpaceAr(EmoSpace):
    """Arabic text model or classifier based on a Emoji Space"""

    @staticmethod
    def model_fname():
        import EvoMSA
        return 'emo-v%s-ar.evoemo' % EvoMSA.__version__[:3]

    @classmethod
    def create_space(cls, fname, output=None, lang='ar', **kwargs):
        super(EmoSpaceAr, cls).create_space(fname, output=output, lang=lang, **kwargs)


class Corpus(BaseTextModel):
    """Text model using only words"""

    def __init__(self, corpus=None, **kwargs):
        self._text = os.getenv('TEXT', default='text')
        self._m = {}
        self._num_terms = 0
        self._training = True
        self._textModel = TextModel([''], token_list=[-1])
        if corpus is not None:
            self.fit(corpus)

    def get_text(self, text):
        return text[self._text]

    def fit(self, c):
        [self.__getitem__(x) for x in c]
        self._training = False
        return self

    @property
    def num_terms(self):
        return self._num_terms

    def tokenize(self, text):
        if isinstance(text, dict):
            text = self.get_text(text)
        if isinstance(text, (list, tuple)):
            tokens = []
            for _text in text:
                tokens.extend(self._textModel.tokenize(_text))
            return tokens
        else:
            return self._textModel.tokenize(text)

    def transform(self, texts):
        """Convert test into a vector

        :param texts: List of text to be transformed
        :type text: list

        :rtype: list

        Example:

        >>> from microtc.textmodel import TextModel
        >>> corpus = ['buenos dias catedras', 'catedras conacyt']
        >>> textmodel = TextModel().fit(corpus)
        >>> X = textmodel.transform(corpus)
        """
        return self._textModel.tonp([self.__getitem__(x) for x in texts])

    def __getitem__(self, d):
        tokens = []
        for t in self.tokenize(d):
            try:
                index, k = self._m[t]
                if self._training:
                    self._m[t] = [index, k+1]
            except KeyError:
                if not self._training:
                    continue
                index, k = self._num_terms, 1
                self._m[t] = [index, k]
                self._num_terms += 1
            tokens.append([index, k])
        return tokens


class AggressivenessAr(Corpus):
    """Arabic text model using an aggressive corpus"""

    def __init__(self, *args, **kwargs):
        fname = os.path.join(os.path.dirname(__file__), 'conf', 'aggressiveness.ar')
        corpus = []
        for x in tweet_iterator(fname):
            corpus += x['words']
        super(AggressivenessAr, self).__init__(corpus)


class AggressivenessEn(Corpus):
    """English text model using an aggressive corpus"""

    def __init__(self, *args, **kwargs):
        fname = os.path.join(os.path.dirname(__file__), 'conf', 'aggressiveness.en')
        corpus = []
        for x in tweet_iterator(fname):
            corpus += x['words']
        super(AggressivenessEn, self).__init__(corpus)


class AggressivenessEs(Corpus):
    """Spanish text model using an aggressive corpus"""

    def __init__(self, *args, **kwargs):
        fname = os.path.join(os.path.dirname(__file__), 'conf', 'aggressiveness.es')
        corpus = []
        for x in tweet_iterator(fname):
            corpus += x['words']
        super(AggressivenessEs, self).__init__(corpus)


class Bernulli(BaseClassifier):
    """Bernulli classifier"""

    def __init__(self, random_state=0):
        self._num_terms = -1

    @property
    def num_terms(self):
        return self._num_terms

    def fit(self, X, klass):
        self._num_terms = X.shape[1]
        klasses = np.unique(klass)
        pr = np.zeros((klasses.shape[0], self.num_terms))
        for i, k in zip(X, klass):
            index = i.indices
            if index.shape[0] > 0:
                pr[k, index] += 1
        _ = np.atleast_2d(self.num_terms + np.array([(klass == _k).sum() for _k in klasses])).T
        pr = (1 + pr) / _
        self._wj = np.log(pr)
        self._vj = np.log(1 - pr)
        return self

    def predict(self, X):
        return np.argmax(self.decision_function_raw(X), axis=1)

    def decision_function(self, X):
        _ = self.predict_proba(X)
        df = _ * 2 - 1
        df[df > 1] = 1
        df[df < -1] = -1
        return df

    def predict_proba(self, X):
        X = self.decision_function_raw(X)
        pr = np.exp(X)
        den = pr.sum(axis=1)
        _ = pr / np.atleast_2d(den).T
        return _

    def decision_function_raw(self, X):
        wj = self._wj
        vj = self._vj
        hy = []
        for d in X:
            x = np.zeros(self.num_terms)
            index = d.indices
            if index.shape[0] > 0:
                x[index] = 1
            _ = ((x * wj) + (1 - x) * vj).sum(axis=1)
            hy.append(_)
        return np.array(hy)


class Multinomial(Bernulli):
    """Multinomial classifier"""

    def fit(self, X, klass):
        self._num_terms = X.shape[1]
        klasses = np.unique(klass)
        pr = np.zeros((klasses.shape[0], self.num_terms))
        for i, k in zip(X, klass):
            index = i.indices
            if index.shape[0] > 0:
                pr[k, index] += 1
        den = pr.sum(axis=1)
        self._log_xj = np.log((1 + pr) / np.atleast_2d(self.num_terms + den).T)
        return self

    def decision_function_raw(self, X):
        xj = self._log_xj
        hy = []
        for d in X:
            x = np.zeros(self.num_terms)
            index = d.indices
            if index.shape[0] > 0:
                x[index] = 1
            _ = (xj * x).sum(axis=1)
            hy.append(_)
        return np.array(hy)


class ThumbsUpDownEs(ThumbsUpDown, BaseTextModel):
    """Spanish thumbs up and down model"""

    def __init__(self, *args, **kwargs):
        super(ThumbsUpDownEs, self).__init__(lang=_SPANISH, stemming=False)

    def fit(self, X):
        return self


class ThumbsUpDownEn(ThumbsUpDown, BaseTextModel):
    """English thumbs up and down model"""

    def __init__(self, *args, **kwargs):
        super(ThumbsUpDownEn, self).__init__(lang=_ENGLISH, stemming=False)

    def fit(self, X):
        return self


class ThumbsUpDownAr(ThumbsUpDown, BaseTextModel):
    """Arabic thumbs up and down model"""

    def __init__(self, *args, **kwargs):
        super(ThumbsUpDownAr, self).__init__(lang=_ARABIC, stemming=False)

    def fit(self, X):
        return self


class Vec(BaseTextModel):
    """Read the key vec, useful to incorporate external knowledge as FastText print-sentence-vectors"""

    def __getitem__(self, x):
        return x['vec']

    def fit(self, X):
        return self


class SemanticToken(BaseTextModel):
    def __init__(self, corpus=None, threshold=0.001, token_min_filter=0.001,
                 token_list=[-2, -1],
                 num_option='delete', usr_option='delete',
                 url_option='delete', emo_option='delete', **kwargs):
        self._text = os.getenv('TEXT', default='text')
        self._textmodel = TextModel(None, token_list=token_list,
                                    threshold=threshold,
                                    token_min_filter=token_min_filter,
                                    num_option=num_option, usr_option=usr_option,
                                    url_option=url_option, emo_option=emo_option,
                                    **kwargs)
        self._threshold = threshold
        if corpus is not None:
            self.fit(corpus)

    def fit(self, X):
        self.init(X)
        return self

    @property
    def threshold(self):
        """Threshold used to remove those tokens that less than 1 - entropy

        :rtype: float
        """

        return self._threshold

    @property
    def textmodel(self):
        """TextModel instance"""

        return self._textmodel

    @property
    def id2token(self):
        """Map from id to token

        :rtype: list"""

        return self._id2token

    def init(self, corpus):
        """Initial model"""

        words = self.tokens(corpus)
        self._weight = np.ones(len(words))
        # key = self.semantic_space._text
        X = self.semantic_space.transform([str(x) for x in words])
        self._kdtree = KDTree(X, metric='manhattan')
        w = self.entropy(self.transform(corpus), corpus, ntokens=X.shape[0])
        w = np.where(w > self.threshold)[0]
        self._kdtree = KDTree(X[w], metric='manhattan')
        self._weight = self._weight[w]
        self._id2token = [words[x] for x in w]
        self.compute_idf(self.transform(corpus))

    def entropy(self, Xt, corpus, ntokens):
        """Compute the entropy"""

        y = [x['klass'] for x in corpus]
        klasses = np.unique(y)
        nklasses = klasses.shape[0]
        weight = np.zeros((klasses.shape[0], ntokens))
        for ki, klass in enumerate(klasses):
            for _y, tokens in zip(y, Xt):
                if _y != klass:
                    continue
                for x in np.unique([_[0] for _ in tokens]):
                    weight[ki, x] += 1
        weight = weight / weight.sum(axis=0)
        weight[~np.isfinite(weight)] = 1.0 / nklasses
        logc = np.log2(weight)
        logc[~np.isfinite(logc)] = 0
        if nklasses > 2:
            logc = logc / np.log2(nklasses)
        return (1 + (weight * logc).sum(axis=0))

    def compute_idf(self, Xt):
        N = len(Xt)
        weight = np.zeros_like(self.weight)
        for tokens in Xt:
            for x in np.unique([_[0] for _ in tokens]):
                weight[x] += 1
        self._weight = np.log2(N / weight)

    @property
    def weight(self):
        """weight per token

        :rtype: list"""

        return self._weight

    def tokens(self, corpus):
        """Tokens used for modeling"""
        self.textmodel.fit(corpus)
        return [x for x in self.textmodel.model._w2id.keys()]

    @property
    def semantic_space(self):
        """Semantic space

        :rtype: instance
        """

        try:
            return self._semantic_space
        except AttributeError:
            self._semantic_space = EmoSpaceEs()
        return self._semantic_space

    def get_text(self, text):
        """Return self._text key from text

        :param text: Text
        :type text: dict
        """

        return text[self._text]

    def tokenize(self, text):
        """Tokenize a text
        :param text: Text
        :type text: dict or str

        :rtype: list of str
        """

        textmodel = self.textmodel
        return textmodel.tokenize(text)

    def transform(self, X):
        from collections import Counter
        w = self.weight
        kdtree = self._kdtree
        tokens_doc = [self.tokenize(x) for x in X]
        tokens = []
        [tokens.__iadd__(x) for x in tokens_doc]
        tokens = [x for x in Counter(tokens).keys()]
        # key = self.semantic_space._text
        xx = self.semantic_space.transform(tokens)
        m = {t: (ind[0], w[ind[0]] / (1 + d[0])) for t, d, ind in zip(tokens, *kdtree.query(xx))}
        Xt = []
        for x in tokens_doc:
            unique = {}
            for _ in x:
                r = m[_]
                try:
                    unique[r[0]] += r[1]
                except KeyError:
                    unique[r[0]] = r[1]
            n = np.sqrt(sum([x * x for _, x in unique.items()]))
            Xt.append([(i, x/n) for i, x in unique.items()])
        return Xt


class SemanticTokenEs(SemanticToken):
    def __init__(self, corpus=None, stopwords='delete', **kwargs):
        super(SemanticTokenEs, self).__init__(corpus=corpus, stopwords=stopwords,
                                              lang='es', **kwargs)


class SemanticTokenEn(SemanticToken):
    def __init__(self, corpus=None, del_dup=False, stopwords='delete', **kwargs):
        super(SemanticTokenEn, self).__init__(corpus=corpus, del_dup=del_dup,
                                              stopwords=stopwords,
                                              lang='en', **kwargs)

    @property
    def semantic_space(self):
        """Semantic space

        :rtype: instance
        """

        try:
            return self._semantic_space
        except AttributeError:
            self._semantic_space = EmoSpaceEn()
        return self._semantic_space


class SemanticTokenAr(SemanticToken):
    def __init__(self, corpus=None, stopwords='delete', **kwargs):
        super(SemanticTokenAr, self).__init__(corpus=corpus, stopwords=stopwords,
                                              lang='ar', **kwargs)

    @property
    def semantic_space(self):
        """Semantic space

        :rtype: instance
        """

        try:
            return self._semantic_space
        except AttributeError:
            self._semantic_space = EmoSpaceAr()
        return self._semantic_space


class SemanticAffectiveEs(SemanticTokenEs):
    def tokens(self, corpus):
        """Tokens used for modeling"""
        fname = os.path.join(ConPATH, 'data', 'es.affective.words.json')
        lst = []
        for x in tweet_iterator(fname):
            lst += x['words']
        return lst


class SemanticAffectiveAr(SemanticTokenAr):
    def tokens(self, corpus):
        """Tokens used for modeling"""
        fname = os.path.join(ConPATH, 'data', 'ar.affective.words.json')
        lst = []
        for x in tweet_iterator(fname):
            lst += x['words']
        return lst


class SemanticAffectiveEn(SemanticTokenEn):
    def tokens(self, corpus):
        """Tokens used for modeling"""
        fname = os.path.join(ConPATH, 'data', 'en.affective.words.json')
        lst = []
        for x in tweet_iterator(fname):
            lst += x['words']
        return lst



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from sklearn.preprocessing import LabelEncoder
import os
import hashlib
from urllib import request
from microtc.utils import load_model
import numpy as np


class LabelEncoderWrapper(object):
    """Wrapper of LabelEncoder. The idea is to keep the order when the classes are numbers
    at some point this will help improve the performance in ordinary classification problems

    :param classifier: Specifies whether it is a classification problem
    :type classifier: bool
    """

    def __init__(self, classifier=True):
        self._m = {}
        self._classifier = classifier

    @property
    def classifier(self):
        """Whether EvoMSA is acting as classifier"""

        return self._classifier

    def fit(self, y):
        """Fit the label encoder

        :param y: Independent variables
        :type y: list or np.array
        :rtype: self
        """

        if not self.classifier:
            return self
        try:
            n = [int(x) for x in y]
        except ValueError:
            return LabelEncoder().fit(y)
        self.classes_ = np.unique(n)
        self._m = {v: k for k, v in enumerate(self.classes_)}
        self._inv = {v: k for k, v in self._m.items()}
        return self

    def transform(self, y):
        if not self.classifier:
            return np.array([float(_) for _ in y])
        return np.array([self._m[int(x)] for x in y])

    def inverse_transform(self, y):
        if not self.classifier:
            return y
        return np.array([self._inv[int(x)] for x in y])


class Cache(object):
    """Store the output of the text models"""

    def __init__(self, basename):
        if basename is None:
            self._cache = None
        else:
            dirname = os.path.dirname(basename)
            if len(dirname) and not os.path.isdir(dirname):
                os.mkdir(dirname)
            self._cache = basename

    def __iter__(self):
        if self._cache is None:
            while True:
                yield None
        for i in self.textModels:
            yield i

    @property
    def textModels(self):
        try:
            return self._textModels
        except AttributeError:
            self._textModels = list()
        return self._textModels

    @property
    def ml(self):
        try:
            return self._classifiers
        except AttributeError:
            self._classifiers = list()
        return self._classifiers

    def ml_train(self):
        if self._cache is None:
            while True:
                yield None
        for i in self.ml:
            yield i

    def ml_kfold(self):
        if self._cache is None:
            while True:
                yield None
        for i in self.ml:
            yield i + '-K'

    @staticmethod
    def get_name(value):
        if isinstance(value, str):
            return hashlib.md5(value.encode()).hexdigest()
        else:
            try:
                vv = value.__name__
            except AttributeError:
                vv = value.__class__.__name__
            return vv

    def append(self, value, ml=None):
        if self._cache is None:
            return
        name = self._cache + "-%s" % self.get_name(value)
        if ml is not None:
            self.ml.append(name + '-' + self.get_name(ml))
        self.textModels.append(name)


def download(model_fname):
    if os.path.isfile(model_fname):
        return model_fname
    dirname = os.path.join(os.path.dirname(__file__), 'models')
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    fname = os.path.join(dirname, model_fname)
    if not os.path.isfile(fname):
        request.urlretrieve("http://ingeotec.mx/~mgraffg/models/%s" % model_fname,
                            fname)
    return fname


def get_model(model_fname):
    fname = download(model_fname)
    return load_model(fname)


def linearSVC_array(classifiers):
    """Transform LinearSVC into weight stored in array.array

    :param classifers: List of LinearSVC where each element is binary
    :type classifers: list
    """

    import array
    intercept = array.array('d', [x.intercept_[0] for x in classifiers])
    coef = np.vstack([x.coef_[0] for x in classifiers])
    coef = array.array('d', coef.T.flatten())
    return coef, intercept


def compute_p(syss):
    from scipy.stats import wilcoxon
    p = []
    mu = syss.mean(axis=0)
    best = mu.argmax()
    for i in range(syss.shape[1]):
        if i == best:
            p.append(np.inf)
            continue
        try:
            pv = wilcoxon(syss[:, best], syss[:, i])[1]
            p.append(pv)
        except ValueError:
            p.append(np.inf)
    ps = np.argsort(p)
    alpha = [np.inf for _ in ps]
    m = ps.shape[0] - 1
    for r, i in enumerate(ps[:-1]):
        alpha_c = (0.05 / (m + 1 - (r + 1)))
        if p[i] > alpha_c:
            break
        alpha[i] = alpha_c
    return p, alpha



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from multiprocessing import cpu_count
import importlib
import numpy as np
import logging
from multiprocessing import Pool
from b4msa.command_line import load_json
from microtc.textmodel import TextModel
from b4msa.lang_dependency import get_lang
from sklearn.model_selection import KFold
from .model import Identity, BaseTextModel, EvoMSAWrapper
from .utils import LabelEncoderWrapper, download
from microtc.utils import load_model, save_model
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


def kfold_decision_function(args):
    cl, X, y, tr, ts, seed = args
    try:
        c = cl(random_state=seed)
    except TypeError:
        c = cl()
    if isinstance(X, (list, tuple)):
        c.fit([X[x] for x in tr], [y[x] for x in tr])
        _ = c.decision_function([X[x] for x in ts])
    else:
        y = np.array(y)
        c.fit(X[tr], y[tr])
        _ = c.decision_function(X[ts])
    return ts, _


def transform(args):
    k, cl, X = args
    df = cl.decision_function(X)
    d = [EvoMSA.tolist(_) for _ in df]
    return (k, d)


def vector_space(args):
    k, t, X, output = args
    if output is not None and os.path.isfile(output):
        return k, load_model(output)
    try:
        res = t.transform(X)
    except AttributeError:
        res = t.tonp([t[_] for _ in X])
    if output is not None:
        save_model(res, output)
    return k, res


DEFAULT_CL = dict(fitness_function='macro-F1',
                  random_generations=1000,
                  orthogonal_selection=True)


DEFAULT_R = dict(random_generations=1000,
                 classifier=False,
                 orthogonal_selection=True)


class EvoMSA(object):
    """
    This is the main entry to create an EvoMSA model

    Let us start with an example to show how to create an EvoMSA model.
    The first thing would be to read the dataset,
    EvoMSA has a dummy dataset to test its functionality, so lets used it.

    Read the dataset

    >>> from EvoMSA import base
    >>> from microtc.utils import tweet_iterator
    >>> import os
    >>> tweets = os.path.join(os.path.dirname(base.__file__), 'tests', 'tweets.json')
    >>> D = [[x['text'], x['klass']] for x in tweet_iterator(tweets)]

    Once the dataset is loaded, it is time to create an EvoMSA model

    >>> from EvoMSA.base import EvoMSA
    >>> evo = EvoMSA(n_jobs=-1).fit([x[0] for x in D], [x[1] for x in D])

    Predict a sentence in Spanish

    >>> evo.predict(['EvoMSA esta funcionando'])
    array(['P'], dtype='<U4')

    :param b4msa_args: Arguments pass to TextModel updating the default arguments
    :type b4msa_args:  dict
    :param stacked_method_args: Arguments pass to the stacked method
    :type stacked_method_args: dict
    :param n_jobs: Multiprocessing default 1 process, <= 0 to use all processors
    :type n_jobs: int
    :param n_splits: Number of folds to train EvoDAG or evodag_class
    :type n_splits: int
    :param seed: Seed used default 0
    :type seed: int
    :param classifier: EvoMSA as classifier default True
    :type classifier: bool
    :param models: Models used as list of pairs (see flags: TR, TH and Emo)
    :type models: list
    :param stacked_method: Classifier or regressor used to ensemble the outputs of :attr:`models` default :class:`EvoDAG.model.EvoDAGE`
    :type stacked_method: str or class
    :param TR: Use b4msa.textmodel.TextModel, sklearn.svm.LinearSVC on the training set
    :type TR: bool
    :param Emo: Use EvoMSA.model.EmoSpace[Ar|En|Es], sklearn.svm.LinearSVC
    :type Emo: bool
    :param TH: Use EvoMSA.model.ThumbsUpDown[Ar|En|Es], sklearn.svm.LinearSVC
    :type TH: bool
    :param HA: Use HA datasets, sklearn.svm.LinearSVC
    :type HA: bool
    :param tm_n_jobs: Multiprocessing using on the Text Models, <= 0 to use all processors
    :type tm_n_jobs: int
    :param cache: Store the output of text models
    :type cache: str
    """

    def __init__(self, b4msa_args=dict(), stacked_method_args=dict(), n_jobs=1,
                 n_splits=5, seed=0, classifier=True, models=None, lang=None,
                 stacked_method="EvoDAG.model.EvoDAGE", TR=True, Emo=False,
                 TH=False, HA=False, tm_n_jobs=None, cache=None):
        if models is None:
            models = []
        if TR:
            models.insert(0, ["b4msa.textmodel.TextModel",
                              "sklearn.svm.LinearSVC"])
        lang = lang if lang is None else get_lang(lang)
        b4msa_args['lang'] = lang
        if Emo:
            models.append(["EvoMSA.model.%s" % self._emoSpace(lang),
                           "sklearn.svm.LinearSVC"])
        if TH:
            models.append(["EvoMSA.model.%s" % self._thumbsUpDown(lang),
                           "sklearn.svm.LinearSVC"])
        if HA:
            fname = download("%s.evoha" % lang)
            models.append([fname, "sklearn.svm.LinearSVC"])
        self._b4msa_args = b4msa_args
        self._evodag_args = stacked_method_args
        if stacked_method == "EvoDAG.model.EvoDAGE":
            if classifier:
                _ = DEFAULT_CL.copy()
            else:
                _ = DEFAULT_R.copy()
        else:
            _ = dict()
        _.update(self._evodag_args)
        self._evodag_args = _
        self._n_jobs = n_jobs if n_jobs > 0 else cpu_count()
        self._tm_n_jobs = tm_n_jobs if tm_n_jobs is None or tm_n_jobs > 0 else cpu_count()
        self._n_splits = n_splits
        self._seed = seed
        self._svc_models = None
        self._evodag_model = None
        self._logger = logging.getLogger('EvoMSA')
        self._le = None
        self._classifier = classifier
        self.cache = cache
        self.models = models
        self._evodag_class = self.get_class(stacked_method)

    def _emoSpace(self, lang):
        m = None
        if lang == 'spanish':
            m = "EmoSpaceEs"
        elif lang == 'english':
            m = "EmoSpaceEn"
        elif lang == 'arabic':
            m = "EmoSpaceAr"
        assert m is not None
        return m

    def _thumbsUpDown(self, lang):
        m = None
        if lang == 'spanish':
            m = "ThumbsUpDownEs"
        elif lang == 'english':
            m = "ThumbsUpDownEn"
        elif lang == 'arabic':
            m = "ThumbsUpDownAr"
        assert m is not None
        return m

    def first_stage(self, X, y):
        """Training EvoMSA's first stage
        
        :param X: Independent variables
        :type X: dict or list
        :param y: Dependent variable.
        :type y: list
        :return: List of vector spaces, i.e., second-stage's training set
        :rtype: list

        >>> import os
        >>> from EvoMSA import base
        >>> from microtc.utils import tweet_iterator
        >>> TWEETS = os.path.join(os.path.dirname(__file__), 'tests', 'tweets.json')
        >>> X = [x['text'] for x in tweet_iterator(TWEETS)]
        >>> y = [x['klass'] for x in tweet_iterator(TWEETS)]
        >>> evo = base.EvoMSA()
        >>> D = evo.first_stage(X, y)
        >>> D.shape
        (1000, 4)

        """

        # Instantiate Text Models
        self.model(X)
        # Transform text into a vector space - List of vector spaces
        X_vector_space = self.vector_space(X)
        # Train supervised learning algorithms
        self.fit_svm(X_vector_space, y)
        # KFold to train the stacked_method
        D = self.kfold_supervised_learning(X_vector_space, y)
        return D

    def fit(self, X, y, test_set=None):
        """
        Train the model using a training set or pairs: text,
        dependent variable (e.g., class) EvoMSA is a two-stage procedure;
        the first step is to transform the text into a vector space with
        dimensions related to the number of classes and then
        train a supervised learning algorithm.

        :param X: Independent variables
        :type X: dict or list
        :param y: Dependent variable.
        :type y: list
        :return: EvoMSA instance, i.e., self
        """

        self._le = LabelEncoderWrapper(classifier=self.classifier).fit(y)
        y = self._le.transform(y)
        # Training first stage
        D = self.first_stage(X, y)
        # After the first stage the cache is not needed
        self.cache = None
        # Transform test set to do transductive learning
        if test_set is not None:
            if isinstance(test_set, list):
                test_set = self.transform(test_set)
        # Training stacked_method
        # Start of the second stage
        _ = dict(n_jobs=self.n_jobs, seed=self._seed)
        self._evodag_args.update(_)
        # y = np.array(y)
        try:
            _ = self._evodag_class(**self._evodag_args)
            _.fit(D, y, test_set=test_set)
            self._evodag_model = _
        except TypeError:
            self._evodag_model = self._evodag_class().fit(D, y)
        return self

    @property
    def classifier(self):
        """Whether EvoMSA is acting as classifier"""

        return self._classifier

    def get_class(self, m):
        if isinstance(m, str):
            if os.path.isfile(m):
                return m
            a = m.split('.')
            p = importlib.import_module('.'.join(a[:-1]))
            return getattr(p, a[-1])
        return m

    @property
    def models(self):
        """Models used as list of pairs

        :rtype: list
        """

        return self._models

    @models.setter
    def models(self, models):
        if models is None:
            return
        if not isinstance(models, list):
            models = [models]
        self._models = []
        for m in models:
            if isinstance(m, list):
                textmodel, classifier = m
                tm = self.get_class(textmodel)
                cl = self.get_class(classifier)
            else:
                tm = Identity
                cl = self.get_class(m)
            assert isinstance(tm, str) or (hasattr(tm, 'transform') and hasattr(tm, 'fit'))
            # Initializing the cache
            if self.cache is not None:
                self.cache.append(tm, ml=cl)
            self._models.append([tm, cl])

    @property
    def n_jobs(self):
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, v):
        self._n_jobs = v
        try:
            self._evodag_model._m._n_jobs = v
        except AttributeError:
            pass

    @property
    def tm_n_jobs(self):
        return self._tm_n_jobs

    @tm_n_jobs.setter
    def tm_n_jobs(self, v):
        self._tm_n_jobs = v

    @property
    def textModels(self):
        """Text Models

        :rtype: list
        """

        return self._textModel

    @property
    def cache(self):
        """Basename to store the output of the textmodels"""

        return self._cache

    @cache.setter
    def cache(self, value):
        from .utils import Cache
        self._cache = Cache(value)

    def predict(self, X):
        if self.classifier:
            pr = self.predict_proba(X)
            return self._le.inverse_transform(pr.argmax(axis=1))
        return self.decision_function(X)

    def predict_proba(self, X):
        X = self.transform(X)
        try:
            return self._evodag_model.predict_proba(X)
        except AttributeError:
            index = self._evodag_model.predict(X)
            res = np.zeros((index.shape[0], self._le.classes_.shape[0]))
            res[np.arange(index.shape[0]), index] = 1
            return res

    def raw_decision_function(self, X):
        X = self.transform(X)
        return self._evodag_model.raw_decision_function(X)

    def decision_function(self, X):
        X = self.transform(X)
        return self._evodag_model.decision_function(X)

    def model(self, X):
        m = []
        kwargs = self._b4msa_args
        self._logger.info("Starting TextModel")
        self._logger.info(str(kwargs))
        for tm, cl in self.models:
            if isinstance(tm, str):
                _ = load_model(tm)
                if isinstance(_, EvoMSA):
                    _ = EvoMSAWrapper(evomsa=_)
                m.append(_)
            elif isinstance(tm, type):
                m.append(tm(**kwargs).fit(X))
            else:
                m.append(tm)
        self._textModel = m

    def vector_space(self, X):
        args = [(i, t, X, output) for (i, t), output in zip(enumerate(self.textModels), self.cache)]
        n_jobs = self.n_jobs if self.tm_n_jobs is None else self.tm_n_jobs
        if n_jobs > 1:
            p = Pool(self.n_jobs, maxtasksperchild=1)
            res = [x for x in tqdm(p.imap_unordered(vector_space, args),
                                   total=len(args))]
            res.sort(key=lambda x: x[0])
            p.close()
        else:
            res = [vector_space(x) for x in tqdm(args)]
        return [x[1] for x in res]

    def sklearn_kfold(self, cl, X, y):
        args = []
        klasses = np.unique(y)
        nclass = klasses.shape[0]
        for tr, ts in KFold(n_splits=self._n_splits,
                            shuffle=True, random_state=self._seed).split(X):
            tr_klasses = np.unique([y[i] for i in tr])
            if tr_klasses.shape[0] != nclass:
                for k in klasses:
                    if k not in tr_klasses:
                        candidate = [(i, x) for i, x in enumerate(ts) if y[x] == k][0]
                        tr = tr.tolist()
                        tr.append(candidate[1])
                        tr = np.array(tr)
            args.append([cl, X, y, tr, ts, self._seed])
        return args

    def kfold_decision_function(self, cl, X, y):
        hy = [None for x in y]
        args = self.sklearn_kfold(cl, X, y)
        if self.n_jobs == 1:
            res = [kfold_decision_function(x) for x in tqdm(args,
                                                            total=len(args))]
        else:
            p = Pool(self.n_jobs, maxtasksperchild=1)
            res = [x for x in tqdm(p.imap_unordered(kfold_decision_function,
                                                    args),
                                   total=len(args))]
            p.close()
        for ts, df in res:
            [hy.__setitem__(k, self.tolist(v)) for k, v in zip(ts, df)]
        return hy

    def kfold_supervised_learning(self, X_vector_space, y):
        """KFold to train the stacked_method, i.e., training set

        :rtype: np.array
        """

        D = None
        for (_, cl), Xvs, output in zip(self.models, X_vector_space,
                                        self.cache.ml_kfold()):
            if output is not None and os.path.isfile(output):
                d = load_model(output)
            else:
                d = self.kfold_decision_function(cl, Xvs, y)
                if output is not None:
                    save_model(d, output)
            if D is None:
                D = d
            else:
                [v.__iadd__(w) for v, w in zip(D, d)]
        D = np.array(D)
        D[~np.isfinite(D)] = 0
        return D

    def transform(self, X):
        Xvs = self.vector_space(X)
        args = [(i, cl, X) for (i, cl), X in zip(enumerate(self._svc_models),
                                                 Xvs)]
        n_jobs = self.n_jobs if self.tm_n_jobs is None else self.tm_n_jobs
        if n_jobs > 1:
            p = Pool(n_jobs, maxtasksperchild=1)
            res = [x for x in tqdm(p.imap_unordered(transform, args),
                                   total=len(args))]
            res.sort(key=lambda x: x[0])
            p.close()
        else:
            res = [transform(x) for x in tqdm(args)]
        res = [x[1] for x in res]
        D = res[0]
        [[v.__iadd__(w) for v, w in zip(D, d)] for d in res[1:]]
        _ = np.array(D)
        _[~np.isfinite(_)] = 0
        return _

    def fit_svm(self, Xvs, y):
        svc_models = []
        for (_, cl), X, output in zip(self.models, Xvs, self.cache.ml_train()):
            if output is not None and os.path.isfile(output):
                svc_models.append(load_model(output))
                continue
            try:
                c = cl(random_state=self._seed)
            except TypeError:
                c = cl()
            c.fit(X, y)
            svc_models.append(c)
            if output is not None:
                save_model(c, output)
        self._svc_models = svc_models

    @staticmethod
    def tolist(x):
        if isinstance(x, list):
            return x
        elif isinstance(x, np.ndarray):
            return x.tolist()
        else:
            return [x]

    @staticmethod
    def read_json(fname):
        kw = load_json(fname)
        if isinstance(kw, list):
            kw = kw[0]
        return kw

    def __getstate__(self):
        """Remove attributes unable to pickle"""

        r = self.__dict__.copy()
        try:
            del r['_logger']
        except KeyError:
            pass
        return r



# Copyright 2020 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_cache():
    from EvoMSA.utils import Cache
    import hashlib

    cache = Cache("hola")
    cache.append("bb")
    cache.append(Cache)
    cache.append(Cache(None))
    ll = [hashlib.md5("bb".encode()).hexdigest(), "Cache", "Cache"]
    for i, j in zip(cache, ll):
        print(i)
        assert i == "hola-%s" % j


def test_cache_cl():
    from EvoMSA.utils import Cache
    from sklearn.svm import LinearSVC
    import hashlib

    cache = Cache("hola")
    cache.append("bb", LinearSVC)
    cache.append(Cache, LinearSVC)
    cache.append(Cache(None), LinearSVC)
    ll = [hashlib.md5("bb".encode()).hexdigest(), "Cache", "Cache"]
    for i, j in zip(cache.ml_train(), ll):
        print(i)
        assert i == "hola-%s-%s" % (j, "LinearSVC")

    for i, j in zip(cache.ml_kfold(), ll):
        print(i)
        assert i == "hola-%s-%s-K" % (j, "LinearSVC")
    



# Copyright 2019 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cpython.list cimport PyList_Append
from cpython cimport array

cdef class TextModelPredict:
    cdef public array.array coef
    cdef public array.array intercept
    cdef public Py_ssize_t nklass
    cdef public object textModel

    def __cinit__(self, object textModel, array.array coef, array.array intercept):
        self.textModel = textModel
        self.coef = coef
        self.intercept = intercept
        self.nklass = len(intercept)

    def transform(self, list data, list output):
        cdef str x
        cdef Py_ssize_t k, init, j
        cdef double v
        cdef array.array _output
        cdef double *_output_c
        cdef double *coef_c = self.coef.data.as_doubles
        for x in data:
            _output = array.copy(self.intercept)
            self._transform(x, _output)
            PyList_Append(output, _output)

    cpdef void _transform(self, str x, array.array _output):
        cdef Py_ssize_t k, init, j
        cdef double v
        cdef double *coef_c = self.coef.data.as_doubles
        _output_c = _output.data.as_doubles
        for k, v in self.textModel[x]:
            init = self.nklass * k
            for j in range(self.nklass):
                _output_c[j] = _output_c[j] + coef_c[init + j] * v

    def __getitem__(self, text):
        _output = array.copy(self.intercept)
        self._transform(text, _output)
        return _output



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from EvoMSA.base import EvoMSA
from microtc.utils import tweet_iterator
import os
TWEETS = os.path.join(os.path.dirname(__file__), 'tweets.json')
try:
    from mock import MagicMock
except ImportError:
    from unittest.mock import MagicMock


class StoreDelete(object):
    def __init__(self, func, data, output):
        self._func = func
        self._data = data
        self._output = output
        self._delete = False

    def __enter__(self):
        dirname = os.path.join(get_dirname(), 'models')
        if not os.path.isdir(dirname):
            os.mkdir(dirname)
        self._output = os.path.join(dirname, self._output)
        if not os.path.isfile(self._output):
            self._func(self._data, output=self._output)
            self._delete = True
        return self

    def __exit__(self, *args):
        if self._delete:
            os.unlink(self._output)


def get_data():
    D = [[x['text'], x['klass']] for x in tweet_iterator(TWEETS)]
    X = [x[0] for x in D]
    y = [x[1] for x in D]
    return X, y


def test_TextModel():
    from b4msa.textmodel import TextModel
    X, y = get_data()
    evo = EvoMSA()
    evo.model(X)
    assert isinstance(evo._textModel, list)
    assert isinstance(evo._textModel[0], TextModel)
    assert len(evo._textModel) == 1


def test_vector_space():
    X, y = get_data()
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          n_estimators=3),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']])
    evo.model(X)
    nrows = len(X)
    X = evo.vector_space(X)
    assert X[0].shape[0] == nrows


def test_EvoMSA_kfold_decision_function():
    from sklearn.preprocessing import LabelEncoder
    X, y = get_data()
    le = LabelEncoder().fit(y)
    y = le.transform(y)
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          n_estimators=3),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']])
    evo.model(X)
    X = evo.vector_space(X)
    cl = evo.models[1][1]
    D = evo.kfold_decision_function(cl, X[1], y)
    assert len(D[0]) == 4
    assert isinstance(D[0], list)


def test_EvoMSA_fit():
    from EvoMSA.model import Bernulli
    from EvoDAG.model import EvoDAGE
    from microtc.utils import load_model, save_model
    X, y = get_data()
    print('iniciando')
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          time_limit=5,
                                          n_estimators=5),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                 n_jobs=1).fit(X, y)
    print("Termine fit")
    assert evo
    assert isinstance(evo._svc_models[1], Bernulli)
    assert isinstance(evo._evodag_model, EvoDAGE)
    save_model(evo, 'test.evomodel')
    print("Guarde modelo")
    evo = load_model('test.evomodel')
    print("Cargue modelo")
    assert isinstance(evo._svc_models[1], Bernulli)
    assert isinstance(evo._evodag_model, EvoDAGE)
    os.unlink('test.evomodel')


def test_EvoMSA_predict():
    import numpy as np
    X, y = get_data()
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          time_limit=15, n_estimators=10),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                 n_jobs=1).fit(X, y)
    hy = evo.predict(X)
    assert len(hy) == 1000
    print((np.array(y) == hy).mean(), hy)
    print(evo.predict_proba(X))
    assert (np.array(y) == hy).mean() > 0.8


def test_EvoMSA_predict_proba():
    X, y = get_data()
    evo = EvoMSA(stacked_method_args=dict(popsize=100,
                                          early_stopping_rounds=100,
                                          time_limit=5,
                                          n_estimators=5),
                 n_jobs=2).fit(X, y)
    hy = evo.predict_proba(X)
    assert len(hy) == 1000
    assert hy.min() >= 0 and hy.max() <= 1


def test_binary_labels_json():
    import json
    X, y = get_data()
    h = dict(NONE=0, N=0, NEU=0, P=1)
    y = [h[x] for x in y]
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          time_limit=5,
                                          n_estimators=5),
                 n_jobs=2).fit(X, y)
    hy = evo.predict(X)
    for x in hy:
        print(type(x), str(x))
        _ = json.dumps(dict(klass=str(x)))
    print(_)


def test_EvoMSA_model():
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                   n_jobs=2)
    assert len(model.models) == 2
    model.model(X)
    assert len(model._textModel) == 2
    print(model._textModel)


def test_EvoMSA_fit_svm():
    from sklearn.preprocessing import LabelEncoder
    X, y = get_data()
    from sklearn.svm import LinearSVC
    from EvoMSA.model import Bernulli
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                   n_jobs=2)
    le = LabelEncoder().fit(y)
    y = le.transform(y)
    model.model(X)
    Xvs = model.vector_space(X)
    model.fit_svm(Xvs, y)
    print(model._svc_models)
    assert len(model._svc_models) == 2
    for ins, klass in zip(model._svc_models, [LinearSVC, Bernulli]):
        assert isinstance(ins, klass)


def test_EvoMSA_evodag_class():
    from sklearn.neighbors import NearestCentroid
    import numpy as np
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                   stacked_method="sklearn.neighbors.NearestCentroid", TR=False,
                   n_jobs=2).fit(X, y)
    assert isinstance(model._evodag_model, NearestCentroid)
    cl = model.predict(X)
    hy = model.predict_proba(X)
    cl2 = model._le.inverse_transform(hy.argmax(axis=1))
    print(cl, cl2)
    assert np.all(cl == cl2)


def test_EvoMSA_multinomial():
    from EvoMSA.model import Multinomial
    X, y = get_data()
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          time_limit=5,
                                          n_estimators=5),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Multinomial']],
                 TR=False,
                 n_jobs=1).fit(X, y)
    assert evo
    assert isinstance(evo._svc_models[0], Multinomial)


def test_EvoMSA_empty_string():
    from EvoMSA.model import Multinomial
    X, y = get_data()
    X.append("")
    y.append("NONE")
    evo = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                          time_limit=5,
                                          n_estimators=5),
                 models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Multinomial']],
                 TR=False,
                 n_jobs=1).fit(X, y)
    assert evo
    assert isinstance(evo._svc_models[0], Multinomial)


def test_label_encoder():
    import numpy as np
    from EvoMSA.base import LabelEncoderWrapper
    y = [2, 2, -1, 0, 0]
    l = LabelEncoderWrapper().fit(y)
    print(l._m)
    yy = l.transform(y)
    assert np.all(np.array([2, 2, 0, 1, 1]) == yy)
    y = [x['klass'] for x in tweet_iterator(TWEETS)]
    l = LabelEncoderWrapper().fit(y)


def test_label_encoder_kwargs():
    from EvoMSA.base import LabelEncoderWrapper
    y = [2, 2, -1, 0.3, 0]
    l = LabelEncoderWrapper(classifier=False).fit(y)
    print(l._m)
    yy = l.transform(y)
    assert yy[-2] == 0.3
    yy = l.inverse_transform(y)
    assert yy[-2] == 0.3
    assert not l.classifier


def get_dirname():
    from EvoMSA import base
    import os
    return os.path.dirname(base.__file__)


def test_EvoMSA_regression():
    from EvoMSA.base import LabelEncoderWrapper
    from EvoMSA.model import EmoSpaceEs
    with StoreDelete(EmoSpaceEs.create_space,
                     TWEETS, EmoSpaceEs.model_fname()) as st:
        X, y = get_data()
        X = [dict(text=x) for x in X]
        l = LabelEncoderWrapper().fit(y)
        y = l.transform(y) - 1.5
        evo = EvoMSA(stacked_method_args=dict(popsize=10,
                                              early_stopping_rounds=10,
                                              time_limit=5,
                                              n_estimators=2),
                     classifier=False,
                     models=[['EvoMSA.model.Identity',
                              'EvoMSA.model.EmoSpaceEs']], TR=False,
                     n_jobs=1).fit(X, y)
        assert evo
        df = evo.decision_function(X)
        print(df.shape, df.ndim)
        assert df.shape[0] == len(X) and df.ndim == 1
        df = evo.predict(X)
        assert df.shape[0] == len(X) and df.ndim == 1


def test_EvoMSA_identity():
    from EvoMSA.model import Identity
    import numpy as np
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   models=[['EvoMSA.model.Corpus', 'EvoMSA.model.Bernulli']],
                   TR=False,
                   stacked_method="EvoMSA.model.Identity",
                   n_jobs=2).fit(X, y)
    assert isinstance(model._evodag_model, Identity)
    cl = model.predict(X)
    hy = model.predict_proba(X)
    cl2 = model._le.inverse_transform(hy.argmax(axis=1))
    print(cl, cl2)
    assert np.all(cl == cl2)


def test_EvoMSA_param_TR():
    from EvoMSA.base import EvoMSA
    from b4msa.textmodel import TextModel
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   TR=False, n_jobs=2)
    assert len(model.models) == 0
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   n_jobs=2)
    assert len(model.models) == 1
    print(model.models[0])
    assert model.models[0][0] == TextModel


def test_EvoMSA_param_Emo():
    from EvoMSA.model import EmoSpaceEs, EmoSpaceEn, EmoSpaceAr
    from EvoMSA.base import EvoMSA

    X, y = get_data()
    for cl, lang in zip([EmoSpaceAr, EmoSpaceEn, EmoSpaceEs],
                        ['ar', 'en', 'es']):
        model = EvoMSA(stacked_method_args=dict(popsize=10,
                                                early_stopping_rounds=10,
                                                n_estimators=3),
                       TR=False, lang=lang, Emo=True, n_jobs=2)
        assert len(model.models) == 1
        assert model.models[0][0] == cl


def test_EvoMSA_param_TH():
    from EvoMSA.model import ThumbsUpDownAr, ThumbsUpDownEn, ThumbsUpDownEs
    from EvoMSA.base import EvoMSA

    X, y = get_data()
    for cl, lang in zip([ThumbsUpDownAr, ThumbsUpDownEn, ThumbsUpDownEs],
                        ['ar', 'en', 'es']):
        model = EvoMSA(stacked_method_args=dict(popsize=10,
                                                early_stopping_rounds=10,
                                                n_estimators=3),
                       TR=False, lang=lang, TH=True, n_jobs=2)
        assert len(model.models) == 1
        assert model.models[0][0] == cl


def test_EvoMSA_param_HA():
    from EvoMSA.model import ThumbsUpDownAr, ThumbsUpDownEn, ThumbsUpDownEs
    from EvoMSA.model import HA
    from EvoMSA.base import EvoMSA
    from b4msa.lang_dependency import get_lang
    import os
    X, y = get_data()
    for cl, lang in zip([ThumbsUpDownAr, ThumbsUpDownEn, ThumbsUpDownEs],
                        ['ar', 'en', 'es']):
        with StoreDelete(HA.create_space, TWEETS, "%s.evoha" % get_lang(lang)) as sd:
            model = EvoMSA(stacked_method_args=dict(popsize=10,
                                                    early_stopping_rounds=10,
                                                    n_estimators=3),
                           TR=False, lang=lang, HA=True, n_jobs=2)
            assert len(model.models) == 1
            print(model.models[0][0])
            assert os.path.isfile(model.models[0][0])


def test_EvoMSA_cpu_count():
    from EvoMSA.base import EvoMSA
    from multiprocessing import cpu_count
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                                    n_estimators=3),
                   TR=False, n_jobs=-1)
    print(model.n_jobs, cpu_count())
    assert model.n_jobs == cpu_count()


def test_evomsa_wrapper():
    from microtc.utils import save_model
    from EvoMSA.base import EvoMSA
    from test_base import get_data
    X, y = get_data()
    model = EvoMSA(stacked_method_args=dict(popsize=10,
                                            early_stopping_rounds=10,
                                            n_estimators=3),
                   stacked_method="sklearn.naive_bayes.GaussianNB",
                   n_jobs=2).fit(X, y)
    save_model(model, 'tmp.evomsa')
    assert os.path.isfile('tmp.evomsa')
    evo = EvoMSA(stacked_method_args=dict(popsize=10,
                                          early_stopping_rounds=10,
                                          n_estimators=3),
                 models=[["tmp.evomsa", "EvoMSA.model.Identity"]],
                 stacked_method="sklearn.naive_bayes.GaussianNB",
                 n_jobs=2).fit(X, y)
    assert evo
    os.unlink("tmp.evomsa")


def test_tm_njobs():
    X, y = get_data()
    evo = EvoMSA(tm_n_jobs=2, n_jobs=1, TH=True, lang="es",
                 stacked_method="sklearn.svm.LinearSVC").fit(X, y)
    evo.predict(X)
    assert evo.n_jobs == 1
    assert evo.tm_n_jobs == 2


def test_sklearn_kfold():
    import numpy as np
    evo = EvoMSA(tm_n_jobs=2, n_jobs=1, TH=True, lang="es",
                 n_splits=3, stacked_method="sklearn.svm.LinearSVC")
    D = np.array([0, 1, 1, 1, 2, 2, 2])
    res = evo.sklearn_kfold(None, D, D)
    for _, _, _, tr, ts, _ in res:
        print(tr, ts)
        assert np.unique(D[tr]).shape[0] == 3


def test_model_instance():
    from microtc.textmodel import TextModel
    X, y = get_data()
    tm = TextModel().fit(X)
    evo = EvoMSA(tm_n_jobs=1, n_jobs=1, TR=False, lang="es",
                 models=[[tm, "sklearn.svm.LinearSVC"]],
                 stacked_method="sklearn.svm.LinearSVC").fit(X, y)
    assert evo.models[0][0] == tm


def test_cache():
    import hashlib

    def func(data, output):
        from b4msa.textmodel import TextModel
        from microtc.utils import tweet_iterator, save_model

        tm = TextModel().fit(list(tweet_iterator(data)))
        save_model(tm, output)

    with StoreDelete(func, TWEETS, "textmodel_cache.tm") as sd:
        cache = os.path.join("tm", "train.json")
        evo = EvoMSA(models=[[sd._output, "sklearn.svm.LinearSVC"]],
                     cache=cache)
        assert os.path.isdir("tm")
        output = hashlib.md5(sd._output.encode()).hexdigest()
        output = cache + "-%s" % output
        print(evo.cache.textModels)
        assert evo.cache.textModels[1] == output
        X, y = get_data()
        evo.first_stage(X, y)
        assert os.path.isfile(output)
        ML = list(evo.cache.ml_train())
        ML_K = list(evo.cache.ml_kfold())
        evo = EvoMSA(models=[[sd._output, "sklearn.svm.LinearSVC"]],
                     stacked_method_args=dict(popsize=10,
                                              early_stopping_rounds=10,
                                              n_estimators=3),
                     cache=cache).fit(X, y, test_set=X[:30])
        hy = evo.predict(X[:10])
        print(len(hy), hy)
        assert len(hy) == 10
        for k in ML:
            print(k)
            assert os.path.isfile(k)
        for k in ML_K:
            print(k)
            assert os.path.isfile(k)



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import importlib
import logging
import EvoMSA
from EvoMSA import base
from microtc.utils import save_model
from microtc.utils import tweet_iterator
from sklearn.metrics import f1_score, recall_score, precision_score, accuracy_score
from scipy.stats import pearsonr
import json
import os
import numpy as np
from .utils import compute_p
from multiprocessing import Pool
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


def fitness_vs(k_model):
    k, model = k_model
    if model == '-':
        return k, '-'
    model = CommandLine.load_model(model)
    return k, np.mean([x.fitness_vs for x in model._evodag_model._m.models])


class CommandLine(object):
    def __init__(self):
        self._klass = os.getenv('KLASS', default='klass')
        self._text = os.getenv('TEXT', default='text')
        self._decision_function = os.getenv('DECISION_FUNCTION',
                                            default='decision_function')
        self.parser = argparse.ArgumentParser(description='EvoMSA')
        self._logger = logging.getLogger('EvoMSA')
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='EvoMSA %s' % EvoMSA.__version__)
        pa('--verbose', dest='verbose', default=logging.NOTSET, type=int)
        pa('-n', '--n_jobs', help='Number of cores',
           dest='n_jobs', default=1, type=int)
        pa('-o', '--output-file',
           help='File / directory to store the result(s)', dest='output_file',
           default=None, type=str)

    def parse_args(self):
        self.data = self.parser.parse_args()
        if hasattr(self.data, 'verbose'):
            logging.basicConfig()
            for k in ['EvoDAG', 'EvoMSA']:
                logger = logging.getLogger(k)
                logger.setLevel(self.data.verbose)
                logger.info('Logging to: %s', self.data.verbose)
        self.main()

    @staticmethod
    def get_class(m):
        if isinstance(m, str):
            a = m.split('.')
            p = importlib.import_module('.'.join(a[:-1]))
            return getattr(p, a[-1])
        return m

    @staticmethod
    def load_model(fname):
        from microtc.utils import load_model
        if os.path.isfile(fname):
            return load_model(fname)
        else:
            cls = CommandLine.get_class(fname)
            ins = cls()
            return ins


class CommandLineTrain(CommandLine):
    def __init__(self):
        super(CommandLineTrain, self).__init__()
        self.training_set()
        pa = self.parser.add_argument
        pa('--kw', dest='kwargs', default=None, type=str,
           help='Parameters in json that overwrite EvoMSA default parameters')
        pa('--test_set', dest='test_set', default=None, type=str,
           help='Test set to do transductive learning')

    def training_set(self):
        cdn = 'File containing the training set.'
        pa = self.parser.add_argument
        pa('training_set',  nargs=1,
           default=None, help=cdn)

    def main(self):
        fnames = self.data.training_set
        fname = fnames[0]
        _ = [[x, x[self._klass]] for x in tweet_iterator(fname)]
        D = [x[0] for x in _]
        Y = [x[1] for x in _]
        if self.data.test_set is not None:
            if os.path.isfile(self.data.test_set):
                test_set = [x for x in tweet_iterator(self.data.test_set)]
            else:
                test_set = self.data.test_set
        else:
            test_set = None
        kwargs = dict(n_jobs=self.data.n_jobs)
        if self.data.kwargs is not None:
            _ = json.loads(self.data.kwargs)
            kwargs.update(_)
        evo_kwargs = dict(tmpdir=self.data.output_file + '_dir')
        if "stacked_method_args" in kwargs:
            evo_kwargs.update(kwargs["stacked_method_args"])
            del kwargs["stacked_method_args"]
        evo = base.EvoMSA(stacked_method_args=evo_kwargs, **kwargs)
        evo.fit(D, Y, test_set=test_set)
        save_model(evo, self.data.output_file)


class CommandLinePredict(CommandLine):
    def __init__(self):
        super(CommandLinePredict, self).__init__()
        pa = self.parser.add_argument
        pa('predict_file',  nargs='+',
           default=None, help='File to predict.')
        pa('-m', '--model', dest='model', default=None,
           help='Model')
        pa('--raw-outputs', dest='raw_outputs', default=False,
           action='store_true',
           help='Raw decision function')
        pa('--decision-function', dest='decision_function', default=False,
           action='store_true',
           help='Ensemble decision function')
        pa('--max-lines', dest='max_lines', default=10000, type=int,
           help='Maximum number of lines to predict in an instance')

    def raw_outputs(self, evo, D):
        predict_file = self.data.predict_file[0]
        pr = []
        max_lines = self.data.max_lines
        while len(D):
            pr.append(evo.raw_decision_function(D[:max_lines]))
            del D[:max_lines]
        X = np.concatenate(pr)
        with open(self.data.output_file, 'w') as fpt:
            for x, df in zip(tweet_iterator(predict_file), X):
                _ = {self._decision_function: df.tolist()}
                x.update(_)
                fpt.write(json.dumps(x) + '\n')

    def decision_function(self, evo, D):
        predict_file = self.data.predict_file[0]
        pr = []
        max_lines = self.data.max_lines
        while len(D):
            pr.append(evo.decision_function(D[:max_lines]))
            del D[:max_lines]
        X = np.concatenate(pr)
        with open(self.data.output_file, 'w') as fpt:
            for x, df in zip(tweet_iterator(predict_file), X):
                _ = {self._decision_function: df.tolist()}
                x.update(_)
                fpt.write(json.dumps(x) + '\n')

    def main(self):
        predict_file = self.data.predict_file[0]
        D = [x for x in tweet_iterator(predict_file)]
        evo = self.load_model(self.data.model)
        if self.data.raw_outputs:
            return self.raw_outputs(evo, D)
        elif self.data.decision_function:
            return self.decision_function(evo, D)
        max_lines = self.data.max_lines
        pr = []
        while len(D):
            pr.append(evo.predict_proba(D[:max_lines]))
            del D[:max_lines]
        pr = np.concatenate(pr)
        hy = evo._le.inverse_transform(pr.argmax(axis=1)).tolist()
        with open(self.data.output_file, 'w') as fpt:
            for x, y, df in zip(tweet_iterator(predict_file), hy, pr):
                _ = {self._klass: y, self._decision_function: df.tolist()}
                x.update(_)
                fpt.write(json.dumps(x) + '\n')


class CommandLinePerformance(CommandLine):
    def __init__(self):
        super(CommandLinePerformance, self).__init__()
        g = self.parser.add_mutually_exclusive_group(required=True)
        pa = self.parser.add_argument
        pa('predictions', nargs='*', default=None)
        ga = g.add_argument
        ga('-m', '--model', help='Model(s) - pickle.dump with gzip', dest='model',
           default=None, type=str, nargs='*')
        pa('--score', help='Score - default macroF1', dest='score',
           default='macroF1', type=str)
        ga('-y', help='Output measured', dest='output', default=None, type=str)
        self._macroF1 = lambda x, y: f1_score(x, y, average='macro')
        self._macroRecall = lambda x, y: recall_score(x, y, average='macro')
        self._macroPrecision = lambda x, y: precision_score(x, y, average='macro')
        self._accuracy = lambda x, y: accuracy_score(x, y)
        self._pearsonr = lambda x, y: pearsonr(x, y)[0]

    def output(self):
        y = [x[self._klass] for x in tweet_iterator(self.data.output)]
        le = base.LabelEncoderWrapper().fit(y)
        perf = getattr(self, "_%s" % self.data.score)
        y = le.transform(y)
        D = []
        I = []
        for fname in self.data.predictions:
            if fname == '-':
                D.append(I)
                I = []
                continue
            hy = le.transform([x[self._klass] for x in tweet_iterator(fname)])
            I.append(perf(y, hy))
        if len(I):
            D.append(I)
        D = np.array(D).T
        p, alpha = compute_p(D)
        self._p = p
        self._alpha = alpha
        for _p, _alpha, mu in zip(p, alpha, D.mean(axis=0)):
            cdn = ''
            if np.isfinite(_alpha):
                cdn = " *"
            print("%0.4f" % mu, cdn)

    def main(self):
        if self.data.output is not None:
            return self.output()
        if len([x for x in self.data.model if x == '-']):
            args = [(k, x) for k, x in enumerate(self.data.model)]
            if self.data.n_jobs > 1:
                p = Pool(self.data.n_jobs)
                res = [x for x in tqdm(p.imap_unordered(fitness_vs, args), total=len(args))]
                res.sort(key=lambda x: x[0])
                p.close()
            else:
                res = [fitness_vs(x) for x in tqdm(args)]
            D = []
            I = []
            for _, x in res:
                if x == '-':
                    D.append(I)
                    I = []
                    continue
                I.append(x)
            if len(I):
                D.append(I)
            D = np.array(D).T
        else:
            models = [self.load_model(d) for d in self.data.model]
            D = np.array([[x.fitness_vs for x in m._evodag_model._m.models] for m in models]).T
        # print(D, '***')
        p, alpha = compute_p(D)
        self._p = p
        self._alpha = alpha
        for m, _p, _alpha, mu in zip(self.data.model, p, alpha, D.mean(axis=0)):
            cdn = ''
            if np.isfinite(_alpha):
                cdn = " *"
            print("%0.4f" % mu, m, cdn)


def train(output=False):
    c = CommandLineTrain()
    c.parse_args()
    if output:
        return c


def predict(output=False):
    c = CommandLinePredict()
    c.parse_args()
    if output:
        return c


def performance(output=False):
    c = CommandLinePerformance()
    c.parse_args()
    if output:
        return c



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from microtc.utils import tweet_iterator
from test_base import StoreDelete
TWEETS = os.path.join(os.path.dirname(__file__), 'tweets.json')


def test_corpus():
    from EvoMSA.model import Corpus
    c = Corpus([x['text'] for x in tweet_iterator(TWEETS)])
    a = c['hola hola mundo']
    assert len(a) == 3
    assert a[0] == a[1]


def test_bernulli():
    import numpy as np
    from EvoMSA.model import Corpus, Bernulli
    from sklearn.preprocessing import LabelEncoder
    c = Corpus([x['text'] for x in tweet_iterator(TWEETS)])
    X = c.transform([x['text'] for x in tweet_iterator(TWEETS)])
    y = [x['klass'] for x in tweet_iterator(TWEETS)]
    le = LabelEncoder().fit(y)
    y = le.transform(y)
    b = Bernulli()
    b.fit(X, y)
    pr = b.decision_function(X)
    assert pr.shape[0] == 1000 and pr.shape[1] == 4
    assert np.all((pr <= 1) & (pr >= -1))


def test_multinomial():
    import numpy as np
    from EvoMSA.model import Corpus, Multinomial
    from sklearn.preprocessing import LabelEncoder
    c = Corpus([x['text'] for x in tweet_iterator(TWEETS)])
    X = c.transform([x['text'] for x in tweet_iterator(TWEETS)])
    y = [x['klass'] for x in tweet_iterator(TWEETS)]
    le = LabelEncoder().fit(y)
    y = le.transform(y)
    b = Multinomial()
    b.fit(X, y)
    pr = b.decision_function(X)
    print(pr.shape[0], pr, b.num_terms)
    assert pr.shape[0] == 1000 and pr.shape[1] == 4
    assert np.all((pr <= 1) & (pr >= -1))


def test_EmoSpace():
    from EvoMSA.model import EmoSpaceEs

    class EmoTest(EmoSpaceEs):
        @staticmethod
        def model_fname():
            return 'test.evoemo'

    X = [x for x in tweet_iterator(TWEETS)]
    emo = EmoTest()
    Xs = [emo[x] for x in X]
    assert len(Xs) == len(X) and len(Xs[0]) == 4
    assert emo.decision_function(X).shape[1] == 4
    # assert emo.model_fname() == 'emo-v%s-es.evoemo' % EvoMSA.__version__


def test_EmoSpace_create_space():
    from EvoMSA import base
    import os
    from EvoMSA.model import EmoSpaceEs
    dirname = os.path.join(os.path.dirname(base.__file__), 'models')
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    output = os.path.join(dirname, EmoSpaceEs.model_fname())
    if not os.path.isfile(output):
        EmoSpaceEs.create_space(TWEETS, output=output)
    assert os.path.isfile(output)
    print(output)
    X = [x for x in tweet_iterator(TWEETS)]
    emo = EmoSpaceEs()
    Xs = [emo[x] for x in X]
    assert len(Xs) == len(X) and len(Xs[0]) == 4
    assert emo.decision_function(X).shape[1] == 4


def test_EmoSpaceEs():
    import EvoMSA
    from EvoMSA.model import EmoSpaceEs
    emo = EmoSpaceEs
    assert emo.model_fname() == 'emo-v%s-es.evoemo' % EvoMSA.__version__[:3]


def test_EmoSpace_transform():
    from EvoMSA.model import EmoSpaceEs

    class EmoTest(EmoSpaceEs):
        @staticmethod
        def model_fname():
            return 'test.evoemo'

    X = [x for x in tweet_iterator(TWEETS)]
    emo = EmoTest()
    r = emo.transform(X)
    print(len(r), len(X), len(r[0]), 4)
    assert len(r) == len(X) and len(r[0]) == 4


def test_EmoSpaceEn():
    from EvoMSA.model import EmoSpaceEn
    import EvoMSA
    cls = EmoSpaceEn
    assert cls
    assert cls.model_fname() == 'emo-v%s-en.evoemo' % EvoMSA.__version__[:3]


def test_EmoSpaceAr():
    from EvoMSA.model import EmoSpaceAr
    import EvoMSA
    cls = EmoSpaceAr
    assert cls
    assert cls.model_fname() == 'emo-v%s-ar.evoemo' % EvoMSA.__version__[:3]


def test_ThumbsUpDownEs():
    from EvoMSA.model import ThumbsUpDownEs
    thumbs = ThumbsUpDownEs()
    _ = thumbs['adoracion XxX fervor vergazo']
    assert _ == (2, 1)


def test_ThumbsUpDownEn():
    from EvoMSA.model import ThumbsUpDownEn
    aff = ThumbsUpDownEn()
    _ = aff['adorably XxX elation vergazo']
    assert (2, 0) == _
    print(_)


def test_ThumbsUpDownAr():
    from EvoMSA.model import ThumbsUpDownAr
    aff = ThumbsUpDownAr()
    _ = aff['adorably XxX elation vergazo']
    assert (0, 0) == _


def test_OutputClassifier():
    from EvoMSA.model import Corpus, OutputClassifier
    from sklearn.preprocessing import LabelEncoder
    c = Corpus([x['text'] for x in tweet_iterator(TWEETS)])
    X = c.transform([x['text'] for x in tweet_iterator(TWEETS)])
    y = [x['klass'] for x in tweet_iterator(TWEETS)]
    le = LabelEncoder().fit(y)
    y = le.transform(y)
    b = OutputClassifier(output='xx')
    assert b._output == 'xx'
    b.fit(X, y)
    assert os.path.isfile('xx_train.csv')
    pr = b.decision_function(X)
    assert os.path.isfile('xx_test.csv')
    assert len(open('xx_test.csv').readlines()) == pr.shape[0]
    os.unlink('xx_train.csv')
    os.unlink('xx_test.csv')


def test_AggressivenessEs():
    from EvoMSA.model import AggressivenessEs
    aff = AggressivenessEs()
    _ = aff['indio XxX fervor vergazo']
    assert len(_) == 1


def test_AggressivenessEn():
    from EvoMSA.model import AggressivenessEn
    aff = AggressivenessEn()
    _ = aff['cockhead cockjockey adoracion XxX fervor vergazo']
    print(_)
    assert len(_) == 2


def test_AggressivenessAr():
    from EvoMSA.model import AggressivenessAr
    aff = AggressivenessAr()
    _ = aff['adoracion XxX fervor vergazo']
    assert len(_) == 0


def test_Vec():
    from EvoMSA.model import Vec
    a = dict(vec=[1, 3, 1])
    vec = Vec()
    assert vec[a] == [1, 3, 1]


def test_semantic_token():
    from EvoMSA.model import SemanticTokenEs, EmoSpaceEs

    class EmoTest(EmoSpaceEs):
        @staticmethod
        def model_fname():
            return 'test.evoemo'

    class STest(SemanticTokenEs):
        @property
        def semantic_space(self):
            """Semantic space

            :rtype: instance
            """

            try:
                return self._semantic_space
            except AttributeError:
                self._semantic_space = EmoTest()
            return self._semantic_space

    corpus = [x for x in tweet_iterator(TWEETS)]
    semantic = STest(corpus)
    print(semantic._weight.shape[0])
    assert semantic._weight.shape[0] == 995
    tr = semantic.transform([dict(text='buenos dias')])[0]
    print(tr)
    assert len(tr) == 3
    print([semantic.id2token[x[0]] for x in tr])


def test_semantic_affective_es():
    from EvoMSA.model import SemanticAffectiveEs, EmoSpaceEs

    class EmoTest(EmoSpaceEs):
        @staticmethod
        def model_fname():
            return 'test.evoemo'

    class STest(SemanticAffectiveEs):
        @property
        def semantic_space(self):
            """Semantic space

            :rtype: instance
            """

            try:
                return self._semantic_space
            except AttributeError:
                self._semantic_space = EmoTest()
            return self._semantic_space
    
    corpus = [x for x in tweet_iterator(TWEETS)]
    semantic = STest(corpus)
    tokens = semantic.tokens(None)
    assert tokens
    print(semantic._weight.shape[0])
    assert semantic._weight.shape[0] == 1386


def test_HA():
    from EvoMSA.model import HA
    from EvoMSA.base import EvoMSA
    import os
    X = [x for x in tweet_iterator(TWEETS)]
    HA.create_space(TWEETS, 'ha.model')
    EvoMSA(stacked_method_args=dict(popsize=10, early_stopping_rounds=10,
                            n_estimators=3),
           models=[['ha.model', 'sklearn.svm.LinearSVC']],
           n_jobs=2).fit(X, [x['klass'] for x in X])
    os.unlink('ha.model')


def test_EmoSpace_model_cl():
    from EvoMSA.model import EmoSpace
    emo = EmoSpace(model_cl=EmoSpace._create_space(TWEETS))
    r = emo[dict(text='buenos dias')]
    print(r)
    assert len(r) == 4


def test_LabeledDataSet():
    from EvoMSA.model import LabeledDataSet
    import os
    LabeledDataSet.create_space(TWEETS, 'lb.model')
    assert os.path.isfile('lb.model')
    os.unlink('lb.model')


def test_projection():
    from EvoMSA.model import LabeledDataSet
    from EvoMSA.model import Projection
    from EvoMSA.utils import download
    from microtc.utils import load_model, tweet_iterator
    import numpy as np
    D = [x for x in tweet_iterator(TWEETS)]
    with StoreDelete(LabeledDataSet.create_space, TWEETS, 'lb.model') as sd:
        model = load_model(download('lb.model'))
        pr = np.diag([1, 1, 1, 1])
        projection = Projection(textModel=model, projection=pr)
        m = projection.transform(D)
        m1 = model.transform(D)
        print(m == m1)
        assert np.all(m == m1)



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup
import EvoMSA
import numpy as np
from Cython.Build import cythonize
from setuptools import Extension
import sys
version = EvoMSA.__version__

extension = [Extension('EvoMSA.cython_utils', ["EvoMSA/cython_utils.pyx"],
                       include_dirs=[np.get_include()])]

with open('README.rst') as fpt:
    long_desc = fpt.read()

setup(
    name="EvoMSA",
    description="""Sentiment Analysis System based on B4MSA and EvoDAG""",
    long_description=long_desc,
    version=version,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        'Programming Language :: Python :: 3',
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],
    url='https://github.com/ingeotec/EvoMSA',
    author="Mario Graff",
    author_email="mgraffg@ieee.org",
    packages=['EvoMSA', 'EvoMSA/tests'],
    include_package_data=True,
    zip_safe=False,
    ext_modules=cythonize(extension,
                          compiler_directives={'language_level': sys.version_info[0],
                                               'profile': False,
                                               'nonecheck': False,
                                               'boundscheck': False}),
    package_data={'EvoMSA/tests': ['tweets.json'],
                  '': ['*.pxd']},
    install_requires=['B4MSA', 'ConceptModelling', 'EvoDAG'],
    entry_points={
        'console_scripts': ['EvoMSA-train=EvoMSA.command_line:train',
                            'EvoMSA-predict=EvoMSA.command_line:predict',
                            'EvoMSA-performance=EvoMSA.command_line:performance']
    }
)



# Copyright 2019 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from test_base import StoreDelete, TWEETS
from EvoMSA.model import LabeledDataSet


def labeledDataSet(*args, output=None):
    return LabeledDataSet.create_space(*args, output=output,
                                       emo_option='delete',
                                       token_list=[-3, -1, 4])


def _read_words(lang):
    """Read the words from our aggressive lexicon

    :param lang: Language
    :type lang: str [ar|en|es]
    """

    import os
    from microtc.utils import tweet_iterator
    from EvoMSA import base
    fname = os.path.join(os.path.dirname(base.__file__), 'conf', 'aggressiveness.%s' % lang)
    corpus = []
    for x in tweet_iterator(fname):
        corpus += x['words']
    return corpus


def test_projection():
    from EvoMSA.align import projection
    from EvoMSA.utils import download
    with StoreDelete(LabeledDataSet.create_space, TWEETS, 'emo-static-es.evoemo') as sd2:
        with StoreDelete(labeledDataSet, TWEETS, 'emo-static-en.evoemo') as sd1:
            res = projection(download('emo-static-es.evoemo'),
                             download('emo-static-en.evoemo'),
                             _read_words('es'),
                             _read_words('en'))
            assert res.ndim == 2
            assert res.shape[0] == res.shape[1]



# Copyright 2019 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


def projection(model_from, model_to, text_from, text_to):
    """
    Compute the coefficients to project the output of a Emoji Space in the origin language to the objetive language

    :param lang_from: Origin model
    :type lang_from: str
    :param lang_to: Objective model
    :type lang_to: str [ar|en|es]
    :param text_from: Text in the origin language
    :type text_from: list
    :param text_from: Text in the objective language
    :type text_from: list
    """

    from microtc.utils import load_model
    import numpy as np
    from sklearn.neighbors import KDTree
    model_from = load_model(model_from)
    model_to = load_model(model_to)
    vec_from = model_from.transform(text_from)
    vec_to = model_to.transform(text_to)
    done = set()
    output = []
    X = []
    kdtree = KDTree(vec_to, metric='manhattan')
    ss = kdtree.query(vec_from)[1].flatten()
    for k, j in tqdm(enumerate(ss)):
        if j in done:
            continue
        X.append(vec_from[k])
        output.append(vec_to[j])
        done.add(j)
    output = np.stack(output)
    X = np.stack(X)
    return np.linalg.lstsq(X, output, rcond=None)[0]

